# -*- coding:utf-8 -*-
# 文件:  errcode.py

ERR_DATE_FMT = 200
ERR_DATE_FMT_MSG = '“%s”格式错误【%s】，期望 年-月-日！'
ERR_TIME_FMT = 201
ERR_VERIFY_NEED_PARAM = 500     # 缺少参数
ERR_VERIFY_VAL_EXCEED = 501     # 数值越限
ERR_PUB_LEN = 700
ERR_PUB_LEN_MSG = '“%s”不能超过“%s”字！'
ERR_PUB_NUM = 706
ERR_PUB_NUM_MSG = '【%s】格式错误，请输入数值！'
ERR_PUB_NUM_INT = 707
ERR_PUB_NUM_INT_MSG = '【%s】格式错误，请输入整数！'
ERR_PUB_NUM_PT2 = 708
ERR_PUB_NUM_PT2_MSG = '【%s】最多两位小数！'
ERR_PUB_NEED = 709
ERR_PUB_NEED_MSG = '请输入【%s】!'
ERR_PUB_GT_EQ_0 = 713
ERR_PUB_GT_EQ_0_MSG = '“%s”需大于或等于0！'
ERR_PUB_CUSTOMIZED_ERR = 819
ERR_PUB_CUSTOMIZED_ERR_MSG = '自定义错误提示，通过ppt设置'
ERR_VERIFY_ERROR = 900          # 校验出错
ERR_VERIFY_WHO = 901            # 未知的校验类型
ERR_SYNTAX = 902
ERR_SYNTAX_MSG = '语法错误！[%s]'
